# This file handles data operations:
# * reading from disk,
# * string cleanup,
# * tokenization,
# * 3rd party parsing (stanford corenlp, spacy)
# The objective is to create a single function that preformats data
# This data will later be segmented in Idea Units

import csv
import json
import re
import unicodedata as ud
import spacy
from spacy.tokens import Doc, Token
from .utils import iu_pprint
# Spacy Token Extension
Token.set_extension("iu_index", default=-1, force=True)

#from pathlib import Path
nlp = spacy.load("en_core_web_lg")
# dep_parser = CoreNLPDependencyParser(url="http://localhost:9000")
# gen_parser = CoreNLPParser(url="http://localhost:9000")
ACCEPTABLE_MODELS = ["spacy", "corenlp_dep", "corenlp_ps"]

def clean_str(s):
    ''' string cleanup function '''
    #remove double spaces and newlines/tabs
    space_undoubler = lambda s : re.sub(r'\s\s*',' ', s)

    res = s
    res = res.replace("\s\\t\s", " ")
    res = res.replace("\s\\n\s", " ")
    res = res.replace("\s\\r\s", " ")
    res = res.replace("\s\\f\s", " ")
    res = res.replace("’", "'")
    res = res.replace("“", "\"")
    res = res.replace("”", "\"")
    res = res.replace("``", "\"")
    res = res.replace("''", "\"")
    res = space_undoubler(res)
    #res = re.sub("\s+", " ", res) # replace multiple spaces with a single one
    # ensure that each open parens has at most one whitespace before
    res = re.sub("\s*\\(", " (", res)
    # ensure that each close parens has at most one whitespace afterwards
    res = re.sub("\\)\s*", ") ", res)
    #uncomment to ensure compatibility with segbot
    '''
    res = re.sub("\s*\\.", " .", res) 
    res = re.sub("\\.\s*", ". ", res)
    '''

    #Remove weird unicode chars (emojis)
    res = ''.join([c for c in res if ud.category(c)[0] != 'C'])
    # undouble spaces again
    res = space_undoubler(res)
    #remove trailing spaces
    res = res.strip()
    return res


def read_file(filename):
    ''' Simple file reader. Output is list of sentences '''
    # read file from disk
    lines = []
    with open(filename) as file:
        # Tokenize sentences
        for row in file.readlines():
            # Skip empty lines and start lines
            cleaned_row = clean_str(row)
            if cleaned_row != "" and cleaned_row != "start":
                lines.append(cleaned_row)
    joined_lines = " ".join(lines)
    sents = [sent.text for sent in nlp(joined_lines).sents]
    return sents


def read_buffer(fileBuffer):
    ''' Simple buffer reader. Output is a list of sentences '''
    # read file from disk
    lines = []
    byteString = fileBuffer.read()
    decodedString = byteString.decode('utf-8')
    # Tokenize sentences
    for row in decodedString.splitlines():
        # Skip empty lines and start lines
        cleaned_row = clean_str(row)
        if cleaned_row != "" and cleaned_row != "start":
            lines.append(cleaned_row)
    fileBuffer.close()
    joined_lines = " ".join(lines)
    sents = [sent.text for sent in nlp(joined_lines).sents]
    return sents


def read_filter(file):
    ''' Simple reader for files containing filters '''
    # read file from disk
    lines = []
    if isinstance(file, str):
        # file is a path
        with open(file) as f:
            # Tokenize sentences
            for row in f.readlines():
                # Skip empty lines and start lines
                cleaned_row = clean_str(row)
                if cleaned_row != "" and cleaned_row != "start":
                    lines.append(cleaned_row.lower())
    else:
        # file is a reader
        s = file.decode("utf-8")
        for row in s.split("\n"):
            # Skip empty lines and start lines
            cleaned_row = clean_str(row)
            if cleaned_row != "" and cleaned_row != "start":
                lines.append(cleaned_row.lower())
    return lines


def parse_file(sents, input_models=["spacy"]):
    '''
    File parser
    Accepted models:
    "spacy" : spacy parsers
    "corenlp_dep": Stanford CoreNLP dependency parser
    "corenlp_ps": Stanford CoreNLP ps rule parser
    WARNING: Currently only spacy is supported
    '''
    # filter models with acceptable ones
    models = [model for model in input_models if model in ACCEPTABLE_MODELS]
    # instantiate a dict with each model as a key
    res = {}
    res["raw"] = sents  # add raw file as well
    for model in models:
        res[model] = []

    for sent in sents:
        for model in models:
            parsed_sent = None
            if model == "spacy":
                parsed_sent = nlp(sent.strip())
            '''
            # Only spacy is supported, as it is faster
            elif model == "corenlp_dep":
                parsed_sent, = dep_parser.raw_parse(sent)
            elif model == "corenlp_ps":
                parsed_sent, = gen_parser.raw_parse(sent)
            '''
            res[model].extend(parsed_sent.sents)
    return res


def import_file(f, models=["spacy"]):
    '''
    Wrapper AIO function to import a file.
    It will read from disk, perform cleanup and parse
    The @models argument accepts a list containing the parse models that 
    the user wishes to adopt
    '''
    raws = None
    if isinstance(f, str):
        raws = read_file(f)
    else:
        raws = read_buffer(f)
    file = parse_file(raws, models)
    return file


def retrieve_filenames(namefile, folder):
    '''
    This function retrieves all the filenames listed in ./data/names.txt
    this is to allow flexibility with the amount of files and avoid uploading
    sensitive data (like filenames) on VCS
    '''
    names = []
    sources = []
    sourceSection = False  # bool flag to check if I am in the source section
    with open(namefile) as file:
        rows = file.readlines()
        for row in rows:
            if row[0] == "*":
                sourceSection = True
            elif row[0] != "#":
                if sourceSection:
                    sources.append(folder + row.strip())
                else:
                    names.append(folder + row.strip())
    # print("filenames")
    # print(names)
    # print(sources)
    return names, sources


def import_all_files(filenames, models=None):
    '''
    Wrapper function that imports, cleans and parses all the files at once
    The @models argument accepts a list containing the parse models that
    the user wishes to adopt
    '''
    files = []
    for filename in filenames:
        try:
            files.append(import_file(filename, models))
        except Exception:
            print("{} not found. Skipping...".format(filename))
    return files


def gen_iu_collection(sentences, gold=False):
    '''
    This function extracts the labeled IUS from a document and keeps track of
    discontinuous Idea Units
    '''
    ius = {}
    disc_ius = set()
    label = lambda x: x._.iu_index
    # look at a different label for gold Ius
    if gold is True:
        label = lambda x: x._.gold_iu_index

    prev_label = None
    for sent in sentences:
        for word in sent:
            if prev_label is None:
                # for the first word initialize the dict entry and temp var
                prev_label = label(word)
                ius[label(word)] = []
            # if the label didn't change from the previous word I can assume
            # that this label already has a dict entry
            if label(word) is prev_label:
                ius[label(word)].append(word)
            # THE LABEL CHANGED!
            # if we don't have the label in the dict then add it and move on
            elif label(word) not in ius:
                ius[label(word)] = []
                ius[label(word)].append(word)
                prev_label = label(word)
            # the label is already in the dict. We have a discontinuous IU
            else:
                ius[label(word)].append(word)
                disc_ius.add(label(word))
                prev_label = label(word)
    return ius, disc_ius


def export_csv(table, filename):
    with open(filename, 'w') as writerFile:
        writer = csv.writer(writerFile, dialect="unix")
        for row in table:
            writer.writerow(row)

def export_labeled_ius(text, filename):
    raw_sents = [iu_pprint(sent, opener="", closer="\n") for sent in text]
    raw = "".join(raw_sents)
    with open(filename, 'w') as file:
        file.write(raw)
    return None

def export_labeled_json(text, filename, doc_name):

    doc_type = "Source text"
    if doc_name != "source":
        doc_type = "Summary text"
    data = prepare_json(text, doc_name, doc_type)
    with open(filename, 'w') as outputfile:
        json.dump(data, outputfile)

def prepare_json(text, doc_name, doc_type):
    data = {}
    data['doc_name'] = doc_name
    data['doc_type'] = doc_type
    data['sents'] = []
    disc_labels = []  # list of discontinuous labels
    idx_list = {}
    last_idx = None

    word_index = 0
    max_iu_index = 0
    cur_iu_index = 0

    for sent in text:
        sent_data = {}
        sent_data['words'] = []
        for token in sent:
            if token._.iu_index != last_idx:
                if token._.iu_index in idx_list.keys():
                    disc_labels.append(token._.iu_index)
                    cur_iu_index = idx_list[token._.iu_index]
                else:
                    max_iu_index = max_iu_index + 1
                    cur_iu_index = max_iu_index
                    idx_list[token._.iu_index] = cur_iu_index
                last_idx = token._.iu_index
            word = {
                'text': token.text,
                'word_index': word_index,
                'iu_index': cur_iu_index,
                'iu_label': token._.iu_index,
                'disc': False
            }
            word_index = word_index + 1
            sent_data['words'].append(word)

        data['sents'].append(sent_data)
    for sent in data['sents']:
        for word in sent["words"]:
            if word['iu_label'] in disc_labels:
                word['disc'] = True
    return data


def prepare_man_seg_json(text):
    seg = text
    if not isinstance(text, Doc):
        seg = nlp(clean_str(seg))

    word_index = 0
    sent_data = {}
    sent_data['words'] = []
    for token in seg:
        word = {
            'text': token.text,
            'word_index': word_index,
            'iu_index': None,
            'iu_label': 'MAN',
            'disc': False
        }
        word_index = word_index + 1
        sent_data['words'].append(word)
    return sent_data


def prepare_man_segs_json(segs, doc_name, doc_type):
    data = {}
    data['doc_name'] = doc_name
    data['doc_type'] = doc_type
    data['sents'] = []

    cur_iu_index = 0
    word_index = 0

    for seg in segs:
        if not isinstance(seg, Doc):
            seg = nlp(clean_str(seg))
        sent_data = {}
        sent_data['words'] = []
        for token in seg:
            word = {
                'text': token.text,
                'word_index': word_index,
                'iu_index': cur_iu_index,
                'iu_label': 'MAN',
                'disc': False
            }
            word_index = word_index + 1
            sent_data['words'].append(word)
        cur_iu_index += 1

        data['sents'].append(sent_data)
    for sent in data['sents']:
        for word in sent["words"]:
            word['disc'] = False
    return data


def export_labeled_json(text, filename, doc_name):
    doc_type = "Source text"
    data = {}
    if doc_name != "source":
        data['doc_type'] = "Summary text"
    data = prepare_json(text, doc_name, doc_type)
    with open(filename, 'w') as outputfile:
        json.dump(data, outputfile)


### This function assignes IU labels to the sents read by data.py
def reat_manual_annotation(filename):

    raws = None
    if isinstance(filename, str):
        raws = read_file(filename)
    else:
        raws = read_buffer(filename)
    file = parse_file(raws, ['spacy'])

    
    gold_iu_index = 0 #gold iu index
    disc_dict = {} #dictionary for disc ius indexes
    sent_idx = 0 #raw sents index
    token_idx = 0 #raw token index

    ## DEBUG:
    #set with the sent indexes of Disc Ius
    disc_ius_set = set()

    for disc_index, iu in ius:
        iu_index = None
        if disc_index is not None:
            #we are examining a discontinuous IU
            if disc_index not in disc_dict.keys():
                #if we don't have our index in the dictionary
                disc_dict[disc_index] = gold_iu_index
                #this increases the IU counter only the first time we find
                #a discontinuous IU
                gold_iu_index +=1
            iu_index = disc_dict[disc_index]
        else:
            #If this IU is not a discontinuous IU then we can avoid the
            #dictionary lookup.
            iu_index = gold_iu_index
            gold_iu_index +=1
        # iterate for each word in the gold idea unit
        for word in iu:
            # get the token row from the raw sents data
            token = sents[sent_idx][token_idx]
            if disc_index is not None:
                disc_ius_set.add(sent_idx)
            #check if the two texts are the same
            if word.text == token.text:
                #assign the IU index
                token._.gold_iu_index = iu_index
                #increase indexes for the exploration of the raw data matrix
                token_idx += 1
                if token_idx == len(sents[sent_idx]):
                    sent_idx += 1
                    token_idx = 0
            else:
                print("***SOMETHING REAL BAD HAPPENED***")
                print("Could not match a word from the raw sentences with the respective Idea Unit.")
                print("word: {} iu: {}".format(word.text,token.text))
                print("token.i: {} token.doc: {}".format(word.i, word.doc))
                print("sent_idx: {} token_idx: {}".format(sent_idx,token_idx))
                print("Please remain calm and call an adult")
                print("-----------")
                raise(Exception("gold"))
    return sents, disc_ius_set


 # This file handles data operations for gold standard files.
# It is a different file from data.py since I don't want to load gold related
# functions in production
# (I.e.: when segmenting raw text I can't lookup manual annotation)
#
# Gold data is used to label the parsed data structure from data.py
# A custom iu_index field will be added to each token from data.py.
# This is because I want dependency trees to be evaluated on full sentences and
# not single Idea Units.

from src.iuextract.data import import_file, clean_str
import spacy
from spacy.tokens import Token
nlp = spacy.load("en_core_web_lg")
import re
from src.iuextract.utils import iu_pprint
# Spacy Token Extension
Token.set_extension("gold_iu_index", default=-1, force=True)

# functions
###this function splits the initial manual index from the discontinuous IUs ###
def split_index_iu(sent):
    match = re.match("(\d+)\|(.*)",sent)
    if match is not None:
        #print(match[1],match[2])
        return (match[1],match[2])
    return (None, sent)

### this function imports the gold file and returns a matrix
### the first column will have the discontinuous IU index
### the second column will have the spacy formatted idea unit
def import_gold(filename):
    res = []
    with open(filename) as goldFile:
        reader = goldFile.readlines()
        for row in reader:
            cleaned_row = clean_str(row)
            ## skipping empty lines
            if cleaned_row != "" and cleaned_row != "start":
                disc_index, raw_iu = split_index_iu(cleaned_row)
                res.append([disc_index, nlp(raw_iu.strip())])
    return res

### Wrapper function to import all goldfiles at once
def import_all_gold_files(filenames):
    files = []
    for filename in filenames:
        try:
            files.append(import_gold(filename))
        except:
            print("{} not found. Skipping...".format(filename))
    return files

### This function assignes gold IU labels to the sents read by data.py
def assign_gold_labels(sents, ius):
    gold_iu_index = 0 #gold iu index
    disc_dict = {} #dictionary for disc ius indexes
    sent_idx = 0 #raw sents index
    token_idx = 0 #raw token index

    ## DEBUG:
    #set with the sent indexes of Disc Ius
    disc_ius_set = set()

    for disc_index, iu in ius:
        iu_index = None
        if disc_index is not None:
            #we are examining a discontinuous IU
            if disc_index not in disc_dict.keys():
                #if we don't have our index in the dictionary
                disc_dict[disc_index] = gold_iu_index
                #this increases the IU counter only the first time we find
                #a discontinuous IU
                gold_iu_index +=1
            iu_index = disc_dict[disc_index]
        else:
            #If this IU is not a discontinuous IU then we can avoid the
            #dictionary lookup.
            iu_index = gold_iu_index
            gold_iu_index +=1
        # iterate for each word in the gold idea unit
        for word in iu:
            # get the token row from the raw sents data
            token = sents[sent_idx][token_idx]
            if disc_index is not None:
                disc_ius_set.add(sent_idx)
            #check if the two texts are the same
            if word.text == token.text:
                #assign the IU index
                token._.gold_iu_index = iu_index
                #increase indexes for the exploration of the raw data matrix
                token_idx += 1
                if token_idx == len(sents[sent_idx]):
                    sent_idx += 1
                    token_idx = 0
            else:
                print("***SOMETHING REAL BAD HAPPENED***")
                print("Could not match a word from the raw sentences with the respective Idea Unit.")
                print("word: {} iu: {}".format(word.text,token.text))
                print("token.i: {} token.doc: {}".format(word.i, word.doc))
                print("sent_idx: {} token_idx: {}".format(sent_idx,token_idx))
                print("sent: {}".format(sents[sent_idx]))
                print("iu: {}".format(iu))
                print("Please remain calm and call an adult")
                print("-----------")
                raise(Exception("gold"))
    return sents, disc_ius_set



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="iuextract",
    version="0.0.5",
    author="Gecchele Marcello",
    author_email="git@gecchele.dev",
    description="Extract Idea Units from strings and files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TT-CL/iuextract",
    
    project_urls={
        "Bug Tracker": "https://github.com/TT-CL/iuextract/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires = [
        'setuptools',
        'spacy>=3.0.0',
    ]
)



from itertools import combinations

available_rules = ["R1", "R2", "R3", "R5", "R6", "R8", "R10"]
## This function prints sentences nicely with their IU numbers in brackets
#  you can customize the brackets by changing the opener and closer parameters
def iu_pprint(sent, gold = False, opener="[",closer="]"):
    texts = [token.text for token in sent]
    indexes = None
    if gold is False:
        indexes = [token._.iu_index for token in sent]
    else:
        indexes = [token._.gold_iu_index for token in sent]
    res = ""
    cur_idx = None
    for i in range(len(texts)):
        #print("i:{}, indexes:{}, cur_idx:{}".format(i,indexes[i],cur_idx))
        if indexes[i] != cur_idx:
            #print(indexes[i], cur_idx)
            cur_idx = indexes[i]
            res += closer+opener+"{}|".format(cur_idx)
        res += "{} ".format(texts[i])
    res += closer     #add final closed bracket ] at the end of the string
    res = res[1:] #crop first closed bracket ] from the beginning of the string
    return res

def init_comb(file):
    rule_combinations = get_rule_combinations()

    for sent in file:
        for word in sent:
            for comb in rule_combinations:
                comb_label = "_".join(comb)
                word._.iu_comb[comb_label] = -1

def get_rule_combinations():
    rule_combinations = []
    for i in range(len(available_rules)):
        rule_combinations.extend(combinations(available_rules, i+1))
    rule_combinations = [list(comb) for comb in rule_combinations]
    return rule_combinations

def get_ius_text(sent):
    res_dict = {}
    for tok in sent:
        if tok._.iu_index not in res_dict.keys():
            res_dict[tok._.iu_index] = [tok]
        else:
            res_dict[tok._.iu_index].append(tok)
    return res_dict



# import re
# from itertools import combinations
# from functools import cmp_to_key
from collections import deque
# from pprint import pprint
# from .iu_utils import iu_pprint
from .data import read_filter
from spacy.tokens import Token
import pkgutil

# initialize the list of filtered IUs from an external file
filter_file = pkgutil.get_data(__name__,"transition_signals.txt")
filtered_ius = read_filter(filter_file)

# Spacy Token Extension for IU combinations
Token.set_extension("iu_comb", default={}, force=True)

available_rules = ["R1", "R2", "R3", "R5", "R6", "R8", "R10"]

def label_ius(file, combination_array=None):
    '''
    Wrapper iu extraction function.
    In input it expects a file parsed with spacy
    No output is expected, as the Idea Units labels will be stored inside
    the spacy tokens
    '''
    s_idx = 0
    for sentence in file:
        root = sentence[0].sent.root
        # print("**Sentence:\n{}".format(sentence))
        # print("*root: {}, POS: {}".format(root.text,root.pos_))

        to_process = tag_nodes(sentence, combination_array=None)
        if len(to_process) == 0:
            # no rule is applicable, segment the full sentence.
            # print("No rule applicable to sent:\n\t{}".format(sentence))
            to_process[root] = ["UNL"]
        to_process = order_nodes_bfs_dict(to_process)

        combination_label = None
        if combination_array is not None:
            combination_label = "_".join(combination_array)

        color_ius(sentence, to_process, s_idx, combination_label)

        #Rule 10
        if combination_array is None or "R10" in combination_array:
            inline_fixes(sentence)
        # print(iu_pprint(sentence))
        # print()
        s_idx = s_idx + 1
    return None


subj = ["nsubj", "nsubjpass", "csubj", "csubjpass"]


def is_V_with_S(word):
    '''
    this function says if rule 1 is applyable to word and all its dependants
    '''
    res = False
    # loop only if the word is a verb
    if word.pos_ == "VERB" or word.pos_ == "AUX":
        # don't split multiple auxiliaries as of Rule 4
        if word.dep_ not in subj:
            for child in word.children:
                # if I find one subject as a dependant
                if child.dep_ in subj:
                    res = True
                    break
    return res


def order_nodes_bfs_dict(nodelist):
    '''
    compute a bfs to find the extraction order of nodes
    '''
    # get the root
    root = list(nodelist.keys())[0].sent.root
    order = []
    q = deque([root])
    # end the loop if I found all the nodes or if I explored the whole tree
    while q and nodelist:
        # pop child and add grandchildren to the queue
        cur_node = q.popleft()
        q.extend(cur_node.children)

        # check node
        if cur_node in nodelist.keys():
            order.append([cur_node, nodelist[cur_node]])
            nodelist.pop(cur_node)
    # filter the unwanted IUs
    for node_arr in order:
        # get the head node of the IU
        node = node_arr[0]
        # generate the IU text
        iu_text = " ".join(t.text for t in node.subtree)
        # if the iu.lowercase is in the filtered list:
        if iu_text.lower().strip(",.") in filtered_ius:
            # print("Filtering IU: {}".format(iu_text))
            order.remove(node_arr)
    order.reverse()  # the extraction order needs to be reversed
    return order


'''
# Bool function for Rule 2
def is_relcl(word):
    res = False
    #enter the loop only if we have a relative clause
    if word.dep_ == "relcl":
        for child in word.children:
            # if we have a pronoun
            if child.dep_ == "mark":
                res = True
                break
    return res
'''


def is_sconj(word):
    ''' bool function for Rule 2 '''
    res = False
    # the sub conjunction is the introducion of a prepositional phrase
    if word.pos_ == "SCONJ" and word.dep_ == "prep":
        res = True
    if word.pos_ == "AUX" or word.pos_ == "VERB":
        for child in word.children:
            # if we have a sconj
            if child.dep_ == "mark" and child.pos_ == "SCONJ":
                res = True
                break
    return res


def is_complementizer(word):
    ''' Rule 2B '''
    res = False
    # enter the loop only if we have a clausal complement
    if word.dep_ == "ccomp":
        for child in word.children:
            # if we have a pronoun
            if child.dep_ == "mark":
                res = True
                break
    return res


def find_PP_PC(sentence):
    '''
    This function finds all Prepositional phrases as Prepositional complements
    It will return all prepostional modifiers
    that are followed by prepostional complements
    '''
    preps = []
    for i in range(len(sentence)):
        if sentence[i].dep_ == "prep":
            # check if I still have more words before the end of the sentence
            if (i+1) < len(sentence):
                # if my preposition is followed by the
                if sentence[i+1].dep_ == "pcomp":
                    preps.append(i)
    return preps


def rule_PP(sentence, s_idx):
    # Prepositional phrases as Prepositional complements are phrases
    # formed by a preposition directly followed by a Prepositional complement
    # I suggest using this instead of the arbitrary 5 word limit
    return None


def is_isolated(sequence):
    '''
    This function evaluates whether a sequence of words is isolated.
    A sequence is isolated if I only have an arc between it and the rest of the
    sentence. either 1 head or 1 child.
    '''
    outside_connections = 0
    for word in sequence:
        # if word does not have the attribute head, it is the root
        if hasattr(word, "head"):
            if word.head not in sequence and word.head.pos_ != 'PUNCT':
                outside_connections += 1
        for child in word.children:
            if child not in sequence and child.pos_ != 'PUNCT':
                outside_connections += 1
    # print(outside_connections)
    return outside_connections == 1


def find_seq_head(sequence):
    ''' given a sequence, returns the head of the subtree '''
    node = sequence[0]
    root = node.sent.root
    if node is not root:
        father = node.head
        while father in sequence:
            node = father
            father = node.head
            # a root father will not have a head
            if father == root:
                if father in sequence:
                    node = father
                break
    return node


def citation_check(node):
    res = False
    if node.text.isdigit():
        # the word is a digit and only child of the appos
        if len([child for child in node.children]) == 0:
            # Citation!
            res = True
    return res


def stopword_check(node):
    res = False
    # print("node: {}".format(node))
    # print("is_stop: {}".format(node.is_stop))
    # print("pos: {}".format(node.pos_))
    if node.is_stop or node.pos_ == "PUNCT":
        res = True
        for child in node.children:
            res = res and stopword_check(child)
    return res


def tag_parens(sentence):
    tag_list = []
    q = []
    word_idx = 0
    for word in sentence:
        if word.text == "(":
            q.append((word, word_idx))
        elif word.text == ")":
            # pop the last (
            open_idx = None
            for el, idx in reversed(q):
                if el.text == "(":
                    q.remove((el, idx))
                    open_idx = idx
                    break
            # I found an open paren
            if open_idx is not None:
                # you can slice a sentence by the 2 indexes to find the substr
                seg_slice = sentence[open_idx+1:word_idx]
                # CHECK IF THIS SEGMENT OF TEXT IS ISOLATED
                # FROM THE REST OF THE SENT
                if is_isolated(seg_slice):
                    slice_head = find_seq_head(seg_slice)
                    if citation_check(slice_head):
                        # print("R3parens - filtered due to citation")
                        pass
                    else:
                        tag_list.append([slice_head, "R3.1"])
        word_idx += 1
    return tag_list


def tag_hyphens(sentence):
    tag_list = []
    q = []
    word_idx = 0
    for word in sentence:
        if word.text == "-":
            # first hyphen
            if len(q) == 0:
                q.append(word_idx)
            else:
                prev_idx = q[len(q)-1]
                seg_slice = sentence[prev_idx+1:word_idx]
                if is_isolated(seg_slice):
                    slice_head = find_seq_head(seg_slice)
                    if citation_check(slice_head):
                        # print("R3hypen - filtered due to citation")
                        pass
                    else:
                        tag_list.append([slice_head, "R3.1"])
                else:
                    # maybe the next hyphen couple will be a parenthetic clause
                    q.append(word_idx)
        word_idx += 1
    return tag_list


def tag_commas(sentence):
    tag_list = []
    q = []
    word_idx = 0
    for word in sentence:
        # for the first comma I want to check the slice from the beginning of
        # the sentence.
        prev_idx = -1
        # analize slices between the commas AND
        # the last slice between the last slice and the end of the sentence
        is_valid_last_slice = word_idx == len(sentence)-1 and len(q) != 0
        if word.text == "," or is_valid_last_slice:
            # if we have found some commas beforehand
            if len(q) != 0:
                # start from
                prev_idx = q[len(q)-1]
            # I don't include commas in the slices. This is because they have
            # weird dependencies. I will manually fix them with inline_fixes()
            # This has the sideffect of making it easier to find the right iu
            # when the first IU in a sentence is comprised of stopwords.
            # The comma will tell me where to attach the stopword IU.
            seg_slice = sentence[prev_idx+1:word_idx]
            # print("analyzing slice \"{}\"".format(slice))
            if is_isolated(seg_slice):
                slice_head = find_seq_head(seg_slice)
                # RULE 3 EXCEPTIONS
                if citation_check(slice_head):
                    # print("R3comma - filtered due to citation")
                    # print("--- {} ---".format(slice))
                    pass
                elif stopword_check(slice_head):
                    # print("R3B - filtered due to stopword_check")
                    # print("--- {} ---".format(slice))
                    tag_list.append([slice_head, "JOIN"])
                else:
                    # print("Splitting! Rule R3Bc")
                    # print("--- {} ---".format(slice))
                    tag_list.append([slice_head, "R3"])
            # I always want to keep track of the last comma I found.
            q.append(word_idx)
        word_idx += 1
    return tag_list


def is_infinive_clause(word):
    ''' rule 5 and 6 '''
    res = False
    if word.dep_ == "acl" or word.dep_ == "advcl":
        # infinitival
        if word.tag_ == "TO" or word.tag_ == "VB":
            res = True
        # gerund
        elif word.tag_ == "VBG" or word.tag_ == "VBN":
            res = True
        # else:
            # print(word, word.tag_, word.dep_)
    return res


def is_appos(word):
    ''' boolean form of rule 7b '''
    res = False
    if word.dep_ == "appos":
        # check if the apposition is a citation:
        if citation_check(word):
            # Citation!
            # print("R3.2 - filtered due to citation")
            pass
        else:
            res = True
    return res


def is_infinitive_verbal(word):
    ''' boolean for rule 7c '''
    res = False
    # a verb is infinitive
    if word.pos_ == "VERB" and word.dep_ == "xcomp" and word.tag_ == "VB":
        # print("VERBAL!")
        # print("word: {}, head: {}, head.pos: {}".format(word, word.head, word.head.pos_))
        # Check if we have the auxiliar TO
        for child in word.children:
            if child.tag_ == "TO":
                res = True
                break
        '''
        for child in word.children:
            # the verbal is preceded by to
            bool = child.text.lower() == "to"
            # to is an infinitival to
            bool = bool and child.pos_ == "PART"
            # to is an auxiliary
            bool = bool and child.dep_ == "aux"
            if  bool is True:
                res = True
                break
        '''
    return res


def find_long_PP(sent, tagged_nodes):
    def add_node(word, rule):
        if word not in tagged_nodes.keys():
            tagged_nodes[word] = [rule]
        else:
            tagged_nodes[word].append(rule)
    already_marked = []
    # looking from right to left so that I ensure 5 long pps
    # for word in reversed(sent):
    for word in sent:
        # if the word is a prepositional modifier:
        if word not in already_marked:
            # if the pp head is directly dependant on a verb
            if word.dep_ == "prep" and word.head.pos_ in ["AUX", "VERB"]:
                # count all the children that are not already labeled
                visited = []
                q = deque([word])
                lenght = 0
                while len(q) > 0:
                    el = q.popleft()
                    visited.append(el)
                    if el not in tagged_nodes.keys():
                        # add 1 to the word lenght
                        if word.pos_ != "PUNCT":
                            lenght += 1
                        q.extend(el.children)
                # if the pp is long enough
                if lenght >= 5:
                    already_marked.extend(visited)
                    add_node(word, "R8")
    return tagged_nodes


def inline_fixes(sent):
    previous_label = None
    attach_prev = [",", ".", ")", "!", "?", ";"]
    for i in range(len(sent)):
        word = sent[i]
        # PUNCTUATION FIX
        # attach each comma, fullstop, ), ! and ? to the previous word
        if word.text in attach_prev:
            # OOB check
            if i > 0:
                word._.iu_index = sent[i-1]._.iu_index
        elif word.text == "(":
            if i < len(sent):  # OOB check
                word._.iu_index = sent[i+1]._.iu_index
        # conjunctions go with their follwing word
        if word.pos_ == "CCONJ":
            if i < len(sent):  # OOB check
                word._.iu_index = sent[i+1]._.iu_index
        # JOIN FIX
        # we attach meaningless ius (stopwords) to the left.
        # If they are in the initial position, then we attach them to the right
        if word._.iu_index == "JOIN":
            if previous_label is None:
                # find a new label:
                previous_label = "JOIN"
                # go forward until I find a new label, then backtrack
                j = i+1
                # print("scanning right...")
                while previous_label == "JOIN" and j < len(sent):
                    # print(sent[j], sent[j]._.iu_index)
                    previous_label = sent[j]._.iu_index
                    j += 1
                # Now previous_label is correct
            # change the JOIN label
            word._.iu_index = previous_label

        # store every word's label (look left)
        previous_label = word._.iu_index
    return None


def tag_nodes(sentence, combination_array=None):
    '''
    This function tags all nodes that (along with their dependencies)
    need to be segmented.
    '''
    res = {}

    if combination_array is None:
        combination_array = available_rules

    def add_node(word, rule):
        if word not in res.keys():
            res[word] = [rule]
        else:
            res[word].append(rule)
    for word in sentence:
        if "R1" in combination_array:
            if is_V_with_S(word):
                add_node(word, "R1")
                # print("V with S: {}".format(word))
        if "R2" in combination_array:
            if is_sconj(word):
                add_node(word, "R2")
                # print("sconj: {}".format(word))
            if is_complementizer(word):
                add_node(word, "R2")
                # print("complementizer: {}".format(word))
        if "R3" in combination_array:
            if is_appos(word):
                add_node(word, "R3.2")
                # print("appos: {}".format(word))
        if "R5" in combination_array:
            if is_infinive_clause(word):
                add_node(word, "R5")
                # print("acl: {}".format(word))
        if "R6" in combination_array:
            if is_infinitive_verbal(word):
                add_node(word, "R6.2")
                # print("verbal: {}".format(word))
    # Rule 3
    if "R3" in combination_array:
        for tag in tag_parens(sentence):
            word, rule = tag
            add_node(word, rule)
        for tag in tag_hyphens(sentence):
            word, rule = tag
            add_node(word, rule)
        for tag in tag_commas(sentence):
            word, rule = tag
            add_node(word, rule)
    # pprint(res)
    # print("TAGGING LONG PPS")
    # Rule 8
    if "R8" in combination_array:
        res = find_long_PP(sentence, res)
    # pprint(res)
    return res


def color_ius(sentence, to_process, s_idx, combination_label=None):
    '''
    This function colors each word in the sentence according to
    nodes found in the list to_process.
    '''
    iu_idx = 1
    for node, rule_labels in to_process:
        label = "JOIN"
        if "JOIN" not in rule_labels:
            label = "{}-{}-{}".format(s_idx, iu_idx, ",".join(rule_labels))

        iu_idx = iu_idx + 1
        # color words according to Idea Unit following the order
        q = deque([node])
        while q:
            cur_node = q.popleft()
            q.extend(cur_node.children)
            # print(q)

            if combination_label is None:
                #if I don't specify a comb label, default to standard iu index
                if cur_node._.iu_index == -1:
                    # only color unexplored nodes
                    cur_node._.iu_index = label
            else:
                if cur_node._.iu_comb[combination_label] == -1:
                    # only color unexplored nodes
                    cur_node._.iu_comb[combination_label] = label
    return None
